unit uHashItem;
{$INCLUDE 'jedi.inc'}
{$INCLUDE 'vfy.inc'}

{.DEFINE HASHTOHEX}

interface

uses
  SysUtils, Classes, Windows,
  uPCRE, uFiles,
  uCompatibility, uProgress;

type
  THashItem = class;

  EHashItemError = class(EConvertError);

  THashValue = AnsiString;

  THashState = (hsOK, hsNew, hsChanged, hsDeleted, hsCopiedFrom, hsMovedTo);
  THashStates = set of THashState;

  THashItem = class
    private
      fBasePath: string;
      fPath: string;
      fSize: int64;
      fDate: TDateTime;
      fHash: THashValue;
      fState: THashState;
      fRelated: THashItem;
    protected
      function GetAsString: string; virtual;
      procedure SetAsString(const Value: string); virtual;
      function EncodeHash(const HashValue: THashValue): string; virtual;
      function DecodeHash(const StringValue: string): THashValue; virtual;
      function GetFileName: string; virtual;
    public
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure Assign(Source: THashItem); virtual;
      function ReadFromFile(const ABasePath, APath: string; OnProgress: TProgressEvent = nil {$IFDEF CASESENSITIVESUPPORT} ; CaseSensitive: boolean = False {$ENDIF} ): boolean; virtual;
      property BasePath: string read fBasePath write fBasePath;
      property Path: string read fPath write fPath;
      property Size: int64 read fSize write fSize;
      property Date: TDateTime read fDate write fDate;
      property Hash: THashValue read fHash write fHash;
      property State: THashState read fState write fState;
      property Related: THashItem read fRelated write fRelated;
      property FileName: string read GetFileName;
      property AsString: string read GetAsString write SetAsString;
    end;

function HashItemCompareByPath(Item1, Item2: Pointer): integer;
function HashItemCompareByHash(Item1, Item2: Pointer): integer;

implementation

uses
  {$IFDEF DELPHI2009_UP} RTLConsts, {$ELSE} {$IFNDEF FPC} Consts, {$ENDIF} {$ENDIF}
  DECHash, DECUtil, DECFmt;

type
  THash = THash_MD4;
  THashFormat = {$IFDEF HASHTOHEX} TFormat_Binary {$ELSE} TFormat_MIME64 {$ENDIF};

{$IFDEF FPC}
const SFOpenError = 'Cannot open file %s';
{$ENDIF}

const READ_BUFFER_SIZE = 8 * 1024 * 1024;

var ReadBuffer: Pointer;

{$IFDEF LARGEFILES}
const READ_BUFFER_ALIGN = 65536;

var ReadBufferAllocated: Pointer;
{$ENDIF}

var ItemAsStringRegExp: TRegExp = nil;

const
  BLANK_CHARS = [#0, #9, #10, #13, ' '];
  HASH_CHARS = {$IFDEF HASHTOHEX} ['0'..'9', 'a'..'f', 'A'..'F'] {$ELSE} ['0'..'9', 'a'..'z', 'A'..'Z', '+', '/', '='] {$ENDIF};
  HASH_CHARS_REGEXP = {$IFDEF HASHTOHEX} '[0-9a-fA-F]' {$ELSE} '[0-9a-zA-Z+/=]' {$ENDIF};

function HashItemCompareByPath(Item1, Item2: Pointer): integer;
var It1, It2: THashItem;
begin
  It1 := THashItem(Item1);
  It2 := THashItem(Item2);
  Result := AnsiCompareText(It1.Path, It2.Path);
end;

function HashItemCompareByHash(Item1, Item2: Pointer): integer;
var It1, It2: THashItem;
begin
  It1 := THashItem(Item1);
  It2 := THashItem(Item2);
  if It1.Hash = It2.Hash then
    if It1.Size = It2.Size then
      Result := 0
    else if It1.Size < It2.Size then
      Result := -1
    else
      Result := 1
  else if It1.Hash < It2.Hash then
    Result := -1
  else
    Result := 1;
end;

{ THashItem }

constructor THashItem.Create;
begin
  inherited Create;
  Clear;
end;

destructor THashItem.Destroy;
begin
  inherited;
end;

procedure THashItem.Clear;
begin
  BasePath := '';
  Path := '';
  Size := -1;
  Date := 0;
  Hash := '';
  State := hsOK;
  Related := nil;
end;

procedure THashItem.Assign(Source: THashItem);
begin
  BasePath := Source.BasePath;
  Path := Source.Path;
  Size := Source.Size;
  Date := Source.Date;
  Hash := Source.Hash;
  State := Source.State;
  Related := Source.Related;
end;

function THashItem.ReadFromFile(const ABasePath, APath: string; OnProgress: TProgressEvent {$IFDEF CASESENSITIVESUPPORT} ; CaseSensitive: boolean {$ENDIF} ): boolean;
{$UNDEF __HASHITEM_WINAPI}
{$IFDEF WIN32}
  {$IFDEF LARGEFILES} {$DEFINE __HASHITEM_WINAPI} {$ENDIF}
  {$IFDEF CASESENSITIVESUPPORT} {$DEFINE __HASHITEM_WINAPI} {$ENDIF}
{$ENDIF}
var TheHash: THash;
    TheFile: THandleStream;
    Count: integer;
    FileSize, TotalSize: int64;
    {$IFDEF __HASHITEM_WINAPI}
    CreateFileFlags: DWORD;
    FileSizeHigh, FileSizeLow: DWORD;
    Handle: THandle;
    {$ENDIF}
begin
  Clear;
  Self.BasePath := ABasePath;
  Self.Path := APath;
  {$IFDEF __HASHITEM_WINAPI}
    CreateFileFlags := FILE_ATTRIBUTE_NORMAL {$IFDEF LARGEFILES} or FILE_FLAG_NO_BUFFERING {$ENDIF} ;
    {$IFDEF CASESENSITIVESUPPORT}
    if CaseSensitive then
      CreateFileFlags := CreateFileFlags or FILE_FLAG_POSIX_SEMANTICS;
    {$ENDIF}
    Handle := CreateFile(PChar(FileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, CreateFileFlags, 0);
    if Handle = INVALID_HANDLE_VALUE then
      {$IFDEF DELPHI2009_UP}
      Raise EFOpenError.CreateResFmt(@SFOpenErrorEx, [ExpandFileName(FileName), SysErrorMessage(GetLastError)]);
      {$ELSE}
        {$IFDEF FPC}
        Raise EFOpenError.CreateFmt(SFOpenError, [FileName]);
        {$ELSE}
        Raise EFOpenError.CreateResFmt(@SFOpenError, [FileName]);
        {$ENDIF}
      {$ENDIF}
    TheFile := THandleStream.Create(Handle);
  {$ELSE}
  TheFile := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  {$ENDIF}
  try
    FileSize := 0;
    {$IFDEF __HASHITEM_WINAPI}
    FileSizeLow := GetFileSize(TheFile.Handle, @FileSizeHigh);
    if (FileSizeLow = $ffffffff) and (GetLastError <> NO_ERROR) then
      Raise EInOutError.Create('Can''t read file size.')
    else
      TotalSize := int64(FileSizeHigh) shl 32 + int64(FileSizeLow);
    {$ELSE}
    TotalSize := TheFile.Size; // Delphi 2009: This doesn't work with THandleStream
    {$ENDIF}
    TheHash := THash.Create;
    try
      TheHash.Init;
      repeat
        Count := TheFile.Read(ReadBuffer^, READ_BUFFER_SIZE);
        if Count > 0 then
          begin
          FileSize := FileSize + Count;
          TheHash.Calc(ReadBuffer^, Count);
          if Assigned(OnProgress) then
            OnProgress(Self, FileSize, TotalSize);
          end;
      until Count = 0;
      TheHash.Done;
      Self.Hash := TheHash.DigestStr(THashFormat);
      Self.Size := TotalSize;
      Self.Date := FileGetDateTime(FileName);
    finally
      TheHash.Free;
      end;
  finally
    TheFile.Free;
    {$IFDEF __HASHITEM_WINAPI}
      CloseHandle(Handle);
    {$ENDIF}
    end;
  Result := Size >= 0;
{$UNDEF __HASHITEM_WINAPI}
end;

function THashItem.GetAsString: string;
var DateAsString: string;
begin
  if Self.Date > 0 then
    DateAsString := FormatDateTime('yyyy"-"mm"-"dd"T"hh":"nn":"ss.zzz', Self.Date)
  else
    DateAsString := '';
  Result := Format('%-80s %s %16s %s', [AnsiQuotedStr(Path, '"'), DateAsString, IntToStr(Size), EncodeHash(Hash)]);
end;

procedure THashItem.SetAsString(const Value: string);
const INVALID_INPUT = 'Invalid AsString input.';
const REGEXP_FILENAME_PART = '"(?P<PATH>[^"]+)"';
      REGEXP_YEAR_PART = '(?P<YEAR>[0-9]{4})';
      REGEXP_MONTH_PART = '(?P<MONTH>[0-9]{2})';
      REGEXP_DAY_PART = '(?P<DAY>[0-9]{2})';
      REGEXP_HOUR_PART = '(?P<HOUR>[0-9]{2})';
      REGEXP_MINUTE_PART = '(?P<MINUTE>[0-9]{2})';
      REGEXP_SECOND_PART = '(?P<SECOND>[0-9]{2})';
      REGEXP_MILLISECOND_PART = '(?P<MILLISECOND>[0-9]{3})';
      REGEXP_DATE_PART = '(?P<DATE>' + REGEXP_YEAR_PART + '-' + REGEXP_MONTH_PART + '-' + REGEXP_DAY_PART + 'T' + REGEXP_HOUR_PART + ':' + REGEXP_MINUTE_PART + ':' + REGEXP_SECOND_PART + '(?:\.' + REGEXP_MILLISECOND_PART + ')?\s+)?';
      REGEXP_SIZE_PART = '(?P<SIZE>-?[0-9]+)';
      REGEXP_HASH_PART = '(?P<HASH>' + HASH_CHARS_REGEXP + '+)';
      REGEXP_ASSTRING = '^\s*' + REGEXP_FILENAME_PART + '\s+' + REGEXP_DATE_PART + REGEXP_SIZE_PART + '\s+' + REGEXP_HASH_PART + '\s*$';
var Dummy: string;
begin
  Clear;
  if ItemAsStringRegExp = nil then
    ItemAsStringRegExp := RegExCreate(REGEXP_ASSTRING, [rcoIgnoreCase, rcoSingleLine]);
  if not ItemAsStringRegExp.Match(Value) then
    Raise EHashItemError.Create(INVALID_INPUT)
  else
    begin
    Self.Path := ItemAsStringRegExp.SubexpressionByName('PATH');
    if ItemAsStringRegExp.SubexpressionByName('DATE', Dummy) and (Dummy <> '') then
      Self.Date := EncodeDate(StrToInt(ItemAsStringRegExp.SubexpressionByName('YEAR')), StrToInt(ItemAsStringRegExp.SubexpressionByName('MONTH')), StrToInt(ItemAsStringRegExp.SubexpressionByName('DAY')))
                 + EncodeTime(StrToInt(ItemAsStringRegExp.SubexpressionByName('HOUR')), StrToInt(ItemAsStringRegExp.SubexpressionByName('MINUTE')), StrToInt(ItemAsStringRegExp.SubexpressionByName('SECOND')), StrToIntDef(ItemAsStringRegExp.SubexpressionByName('MILLISECOND'), 0));
    Self.Size := StrToInt64(ItemAsStringRegExp.SubexpressionByName('SIZE'));
    Self.Hash := THashValue(ItemAsStringRegExp.SubexpressionByName('HASH'));
    end;
end;

function THashItem.EncodeHash(const HashValue: THashValue): string;
{$IFDEF HASHTOHEX}
const HexChars = '0123456789abcdef';
var i, n: integer;
    Code: Byte;
{$ENDIF}
begin
  {$IFDEF HASHTOHEX}
  n := Length(HashValue);
  SetLength(Result, 2*n);
  for i := 1 to n do
    begin
    Code := Byte(HashValue[i]);
    Result[2*i-1] := HexChars[(Code shr 4)+1];
    Result[2*i] := HexChars[(Code and $f)+1];
    end;
  {$ELSE}
  Result := string(HashValue);
  {$ENDIF}
end;

function THashItem.DecodeHash(const StringValue: string): THashValue;
{$IFDEF HASHTOHEX}
{$ENDIF}
begin
  {$IFDEF HASHTOHEX}
  todo
  {$ELSE}
  Result := THashValue(StringValue);
  {$ENDIF}
end;

function THashItem.GetFileName: string;
begin
  Result := BasePath + Path;
end;

initialization
  {$IFDEF LARGEFILES}
  GetMem(ReadBufferAllocated, READ_BUFFER_SIZE + READ_BUFFER_ALIGN);
  ReadBuffer := Pointer(Longword(ReadBufferAllocated) + READ_BUFFER_ALIGN - (Longword(ReadBufferAllocated) mod READ_BUFFER_ALIGN));
  {$ELSE}
  GetMem(ReadBuffer, READ_BUFFER_SIZE);
  {$ENDIF}
  ItemAsStringRegExp := nil;

finalization
  {$IFDEF LARGEFILES}
  FreeMem(ReadBufferAllocated);
  ReadBufferAllocated := nil;
  {$ELSE}
  FreeMem(ReadBuffer);
  {$ENDIF}
  ReadBuffer := nil;
  RegExFreeAndNil(ItemAsStringRegExp);

end.
