unit uHashList;
{$INCLUDE 'jedi.inc'}
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows, Contnrs,
  uHashItem, uProgress, uFiles;

type
  THashList = class(TObjectList)
    private
      function GetItem(Index: integer): THashItem;
      procedure SetItem(Index: integer; Value: THashItem);
    protected
    public
      constructor Create; virtual;
      destructor Destroy; override;
      function Add(Item: THashItem): Integer; virtual;
      function IndexOf(Item: THashItem): Integer; virtual;
      procedure Insert(Index: Integer; Item: THashItem); virtual;
      function Remove(Item: THashItem): Integer; virtual;
      procedure LoadFromFile(const FileName: string; const BasePath: string = ''; const OnProgress: TProgressEvent = nil; const DefaultState: THashState = hsOK); virtual;
      procedure SaveToFile(const FileName: string; const OnProgress: TProgressEvent = nil; const SkipStates: THashStates = []); virtual;
      procedure DeleteByState(State: THashState); virtual;
      procedure SortByPath; virtual;
      function FindByPath(const Path: string; out Index: integer): boolean; virtual; // Requires HashList sorted by Path
      procedure SortByHash; virtual;
      function FindByHash(const Hash: AnsiString; const Size: int64; out Index: integer): boolean; virtual; // Requires HashList sorted by Hash
      property Items[Index: Integer]: THashItem read GetItem write SetItem; default;
    end;

implementation

{ THashList }

constructor THashList.Create;
begin
  inherited Create(True);
end;

destructor THashList.Destroy;
begin
  inherited;
end;

function THashList.GetItem(Index: integer): THashItem;
begin
  Result := THashItem(inherited Items[Index]);
end;

procedure THashList.SetItem(Index: integer; Value: THashItem);
begin
  inherited Items[Index] := Value;
end;

function THashList.Add(Item: THashItem): Integer;
begin
  Result := inherited Add(Item);
end;

function THashList.IndexOf(Item: THashItem): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure THashList.Insert(Index: Integer; Item: THashItem);
begin
  inherited Insert(Index, Item);
end;

function THashList.Remove(Item: THashItem): Integer;
begin
  Result := inherited Remove(Item);
end;

procedure THashList.LoadFromFile(const FileName, BasePath: string; const OnProgress: TProgressEvent; const DefaultState: THashState);
var T: TextFile;
    PT: ^TextFile;
    s: string;
    Item: THashItem;
    Done, Total: int64;
    StdIn: boolean;
begin
  StdIn := FileName = '';
  if StdIn then
    begin
    Total := -1;
    PT := @Input;
    end
  else
    begin
    Total := FileGetSize(FileName);
    AssignFile(T, FileName);
    Reset(T);
    PT := @T;
    end;
  try
    Done := 0;
    while not Eof(PT^) do
      begin
      Readln(PT^, s);
      s := Trim(s);
      if s <> '' then
        begin
        Item := THashItem.Create;
        try
          Item.BasePath := BasePath;
          Item.State := DefaultState;
          Item.AsString := s;
          Add(Item);
        except
          Item.Free;
          Raise;
          end;
        end;
      if Assigned(OnProgress) and (Total >= 0) then
        begin
        Inc(Done, Length(s)+2);
        OnProgress(Self, Done, Total);
        end;
      end;
  finally
    if not StdIn then
      CloseFile(T);
    end;
end;

procedure THashList.SaveToFile(const FileName: string; const OnProgress: TProgressEvent; const SkipStates: THashStates);
var T: TextFile;
    PT: ^TextFile;
    i: integer;
    StdOut: boolean;
begin
  StdOut := FileName = '';
  if StdOut then
    PT := @Output
  else
    begin
    AssignFile(T, FileName);
    Rewrite(T);
    PT := @T;
    end;
  try
    for i := 0 to Pred(Count) do
      begin
      if not (Items[i].State in SkipStates) then
        Writeln(PT^, Items[i].AsString);
      if Assigned(OnProgress) then
        OnProgress(Self, Succ(i), Count);
      end;
  finally
    if not StdOut then
      CloseFile(T);
    end;
end;

procedure THashList.SortByPath;
begin
  inherited Sort(HashItemCompareByPath);
end;

function THashList.FindByPath(const Path: string; out Index: integer): boolean;
var
  LowIndex, HighIndex, TestIndex, CompareResult: Integer;
begin
  Result := False;
  LowIndex := 0;
  HighIndex := Pred(Count);
  while LowIndex <= HighIndex do
    begin
    TestIndex := (LowIndex + HighIndex) shr 1;
    CompareResult := AnsiCompareText(Items[TestIndex].Path, Path);
    if CompareResult < 0 then
      LowIndex := Succ(TestIndex)
    else
      begin
      HighIndex := Pred(TestIndex);
      if CompareResult = 0 then
        Result := True;
      end;
    end;
  Index := LowIndex;
end;

procedure THashList.SortByHash;
begin
  inherited Sort(HashItemCompareByHash);
end;

function THashList.FindByHash(const Hash: AnsiString; const Size: int64; out Index: integer): boolean;
var
  LowIndex, HighIndex, TestIndex, CompareResult: Integer;
begin
  Result := False;
  LowIndex := 0;
  HighIndex := Pred(Count);
  while LowIndex <= HighIndex do
    begin
    TestIndex := (LowIndex + HighIndex) shr 1;
    if Items[TestIndex].Hash < Hash then
      CompareResult := -1
    else if Items[TestIndex].Hash > Hash then
      CompareResult := 1
    else if Items[TestIndex].Size < Size then
      CompareResult := -1
    else if Items[TestIndex].Size > Size then
      CompareResult := 1
    else
      CompareResult := 0;
    if CompareResult < 0 then
      LowIndex := Succ(TestIndex)
    else
      begin
      HighIndex := Pred(TestIndex);
      if CompareResult = 0 then
        Result := True;
      end;
    end;
  Index := LowIndex;
end;

procedure THashList.DeleteByState(State: THashState);
var i: integer;
begin
  for i := Pred(Count) downto 0 do
    if Items[i].State = State then
      Delete(i);
end;

end.
